/*
 * This is a module which is used for rejecting packets.
 * 	Added support for customized reject packets (Jozsef Kadlecsik).
 * Sun 12 Nov 2000
 * 	Port to IPv6 / ip6tables (Harald Welte <laforge@gnumonks.org>)
 */
#include <linux/config.h>
#include <linux/module.h>
#include <linux/skbuff.h>
#include <linux/icmpv6.h>
#include <net/tcp.h>
#include <net/ipv6.h>
#include <net/ip6_fib.h>
#include <net/ip6_route.h>
#include <linux/netfilter_ipv6/ip6_tables.h>
#include <linux/netfilter_ipv6/ip6t_REJECT.h>

#if 0
#define DEBUGP printk
#else
#define DEBUGP(format, args...)
#endif

static int maybe_reroute(struct sk_buff *skb)
{
	if (skb->nfcache & NFC_ALTERED){
		if (ip6_route_me_harder(skb) != 0){
			kfree_skb(skb);
			return -EINVAL;
		}
	}

	return skb->dst->output(skb);
}

/* Send RST reply */
static void send_reset(struct sk_buff *oldskb)
{
	struct sk_buff *nskb;
	struct tcphdr *otcph, *tcph;
	unsigned int otcplen, tcphoff;
	int needs_ack;
	struct ipv6hdr *oip6h = oldskb->nh.ipv6h, *ip6h;
	struct dst_entry *dst;
	u8 proto = oip6h->nexthdr;
	struct flowi fl;

	if ((!(ipv6_addr_type(&oip6h->saddr) & IPV6_ADDR_UNICAST)) ||
	    (!(ipv6_addr_type(&oip6h->daddr) & IPV6_ADDR_UNICAST))) {
		DEBUGP("ip6t_REJECT: addr is not unicast.\n");
		return;
	}

	tcphoff = ipv6_skip_exthdr(oldskb, ((u8*)(oip6h+1) - oldskb->data),
				   &proto, oldskb->len - ((u8*)(oip6h+1)
						       - oldskb->data));

	if ((tcphoff < 0) || (tcphoff > oldskb->len)) {
		DEBUGP("ip6t_REJECT: Can't get TCP header.\n");
		return;
	}

	otcplen = oldskb->len - tcphoff;

	/* IP header checks: fragment, too short. */
	if ((proto != IPPROTO_TCP) || (otcplen < sizeof(struct tcphdr))) {
		DEBUGP("ip6t_REJECT: proto(%d) != IPPROTO_TCP, or too short. otcplen = %d\n",
			proto, otcplen);
		return;
	}

	otcph = (struct tcphdr *)(oldskb->data + tcphoff);

	/* No RST for RST. */
	if (otcph->rst) {
		DEBUGP("ip6t_REJECT: RST is set\n");
		return;
	}

	/* Check checksum. */
	if (csum_ipv6_magic(&oip6h->saddr, &oip6h->daddr, otcplen, IPPROTO_TCP,
			    skb_checksum(oldskb, tcphoff, otcplen, 0))) {
		DEBUGP("ip6t_REJECT: TCP checksum is invalid\n");
		return;
	}

	memset(&fl, 0, sizeof(fl));
	fl.proto = IPPROTO_TCP;
	fl.fl6_src = &oip6h->daddr;
	fl.fl6_dst = &oip6h->saddr;
	fl.uli_u.ports.sport = otcph->dest;
	fl.uli_u.ports.dport = otcph->source;
	dst = ip6_route_output(NULL, &fl);
	if (dst->error) {
		if (net_ratelimit())
			printk("ip6t_REJECT: can't find dst. err = %d\n",
			       dst->error);
		dst_release(dst);
		return;
	}

	nskb = alloc_skb(MAX_HEADER + sizeof(struct tcphdr)
			 + dst->dev->hard_header_len + 15, GFP_ATOMIC);
	if (!nskb) {
		if (net_ratelimit())
			printk("ip6t_REJECT: Can't alloc skb\n");
		dst_release(dst);
		return;
	}

	nskb->dst = dst;

	skb_reserve(nskb, (dst->dev->hard_header_len + 15) & ~15);

	ip6h = nskb->nh.ipv6h = (struct ipv6hdr *)
					skb_put(nskb, sizeof(struct ipv6hdr));
	ip6h->version = 6;
	ip6h->hop_limit = ((struct rt6_info*)dst)->rt6i_hoplimit;
	ip6h->nexthdr = IPPROTO_TCP;
	ip6h->payload_len = htons(sizeof(struct tcphdr));
	ipv6_addr_copy(&ip6h->saddr, &oip6h->daddr);
	ipv6_addr_copy(&ip6h->daddr, &oip6h->saddr);

	tcph = (struct tcphdr *)skb_put(nskb, sizeof(struct tcphdr));
	/* Truncate to length (no data) */
	tcph->doff = sizeof(struct tcphdr)/4;
	tcph->source = otcph->dest;
	tcph->dest = otcph->source;

	if (otcph->ack) {
		needs_ack = 0;
		tcph->seq = otcph->ack_seq;
		tcph->ack_seq = 0;
	} else {
		needs_ack = 1;
		tcph->ack_seq = htonl(ntohl(otcph->seq) + otcph->syn
				+ otcph->fin + otcplen - (otcph->doff<<2));
		tcph->seq = 0;
	}

	/* Reset flags */
	((u_int8_t *)tcph)[13] = 0;
	tcph->rst = 1;
	tcph->ack = needs_ack;
	tcph->window = 0;
	tcph->urg_ptr = 0;
	tcph->check = 0;

	/* Adjust TCP checksum */
	tcph->check = csum_ipv6_magic(&nskb->nh.ipv6h->saddr,
				      &nskb->nh.ipv6h->daddr,
				      sizeof(struct tcphdr), IPPROTO_TCP,
				      csum_partial((char *)tcph,
						   sizeof(struct tcphdr), 0));

	NF_HOOK(PF_INET6, NF_IP6_LOCAL_OUT, nskb, NULL, nskb->dst->dev,
		maybe_reroute);
}

static unsigned int reject6_target(struct sk_buff **pskb,
			   unsigned int hooknum,
			   const struct net_device *in,
			   const struct net_device *out,
			   const void *targinfo,
			   void *userinfo)
{
	const struct ip6t_reject_info *reject = targinfo;
	struct sk_buff *skb2 = NULL;
	struct rt6_info *rt6i;
	struct net_device odev;

	if (!out) {
		skb2 = skb_clone(*pskb, GFP_ATOMIC);
		if (skb2 == NULL) {
			return NF_DROP;
		}
		dst_release(skb2->dst);
		skb2->dst = NULL;

		rt6i = rt6_lookup(&skb2->nh.ipv6h->saddr, NULL, 0, 0);
		if (rt6i && rt6i->rt6i_dev) {
			skb2->dev = rt6i->rt6i_dev;
			rt6i = rt6_lookup(&skb2->nh.ipv6h->daddr, &skb2->nh.ipv6h->saddr, 0, 0);
		}
		memcpy(&odev, skb2->dev, sizeof(odev)); /* XXX 'out' has 'const' qualifier... */
	} else {
		memcpy(&odev, out, sizeof(odev));
	}

	printk(KERN_DEBUG "%s: medium point\n", __FUNCTION__);
	/* WARNING: This code causes reentry within ip6tables.
	   This means that the ip6tables jump stack is now crap.  We
	   must return an absolute verdict. --RR */
	DEBUGP("REJECTv6: calling icmpv6_send\n");
    	switch (reject->with) {
    	case IP6T_ICMP6_NO_ROUTE:
    		icmpv6_send(*pskb, ICMPV6_DEST_UNREACH, ICMPV6_NOROUTE, 0, &odev);
    		break;
    	case IP6T_ICMP6_ADM_PROHIBITED:
    		icmpv6_send(*pskb, ICMPV6_DEST_UNREACH, ICMPV6_ADM_PROHIBITED, 0, &odev);
    		break;
    	case IP6T_ICMP6_NOT_NEIGHBOUR:
    		icmpv6_send(*pskb, ICMPV6_DEST_UNREACH, ICMPV6_NOT_NEIGHBOUR, 0, &odev);
    		break;
    	case IP6T_ICMP6_ADDR_UNREACH:
    		icmpv6_send(*pskb, ICMPV6_DEST_UNREACH, ICMPV6_ADDR_UNREACH, 0, &odev);
    		break;
    	case IP6T_ICMP6_PORT_UNREACH:
    		icmpv6_send(*pskb, ICMPV6_DEST_UNREACH, ICMPV6_PORT_UNREACH, 0, &odev);
    		break;
#if 0
    	case IPT_ICMP_ECHOREPLY: {
		struct icmp6hdr *icmph  = (struct icmphdr *)
			((u_int32_t *)(*pskb)->nh.iph + (*pskb)->nh.iph->ihl);
		unsigned int datalen = (*pskb)->len - (*pskb)->nh.iph->ihl * 4;

		/* Not non-head frags, or truncated */
		if (((ntohs((*pskb)->nh.iph->frag_off) & IP_OFFSET) == 0)
		    && datalen >= 4) {
			/* Usually I don't like cut & pasting code,
			   but dammit, my party is starting in 45
			   mins! --RR */
			struct icmp_bxm icmp_param;

			icmp_param.icmph=*icmph;
			icmp_param.icmph.type=ICMP_ECHOREPLY;
			icmp_param.data_ptr=(icmph+1);
			icmp_param.data_len=datalen;
			icmp_reply(&icmp_param, *pskb);
		}
	}
	break;
#endif
	case IP6T_TCP_RESET:
		send_reset(*pskb);
		break;
	default:
		printk(KERN_WARNING "REJECTv6: case %u not handled yet\n", reject->with);
		break;
	}

	if (skb2) kfree_skb(skb2);

	return NF_DROP;
}

static inline int find_ping_match(const struct ip6t_entry_match *m)
{
	const struct ip6t_icmp *icmpinfo = (const struct ip6t_icmp *)m->data;

	if (strcmp(m->u.kernel.match->name, "icmp6") == 0
	    && icmpinfo->type == ICMPV6_ECHO_REQUEST
	    && !(icmpinfo->invflags & IP6T_ICMP_INV))
		return 1;

	return 0;
}

static int check(const char *tablename,
		 const struct ip6t_entry *e,
		 void *targinfo,
		 unsigned int targinfosize,
		 unsigned int hook_mask)
{
 	const struct ip6t_reject_info *rejinfo = targinfo;

 	if (targinfosize != IP6T_ALIGN(sizeof(struct ip6t_reject_info))) {
  		DEBUGP("REJECTv6: targinfosize %u != 0\n", targinfosize);
  		return 0;
  	}

	/* Only allow these for packet filtering. */
	if (strcmp(tablename, "filter") != 0) {
		DEBUGP("REJECTv6: bad table `%s'.\n", tablename);
		return 0;
	}
	if ((hook_mask & ~((1 << NF_IP6_LOCAL_IN)
			   | (1 << NF_IP6_FORWARD)
			   | (1 << NF_IP6_LOCAL_OUT))) != 0) {
		DEBUGP("REJECTv6: bad hook mask %X\n", hook_mask);
		return 0;
	}

	if (rejinfo->with == IP6T_ICMP6_ECHOREPLY) {
		/* Must specify that it's an ICMP ping packet. */
		if (e->ipv6.proto != IPPROTO_ICMPV6
		    || (e->ipv6.invflags & IP6T_INV_PROTO)) {
			DEBUGP("REJECTv6: ECHOREPLY illegal for non-icmp\n");
			return 0;
		}
		/* Must contain ICMP match. */
		if (IP6T_MATCH_ITERATE(e, find_ping_match) == 0) {
			DEBUGP("REJECTv6: ECHOREPLY illegal for non-ping\n");
			return 0;
		}
	} else if (rejinfo->with == IP6T_TCP_RESET) {
		/* Must specify that it's a TCP packet */
		if (e->ipv6.proto != IPPROTO_TCP
		    || (e->ipv6.invflags & IP6T_INV_PROTO)) {
			DEBUGP("REJECTv6: TCP_RESET illegal for non-tcp\n");
			return 0;
		}
	}

	return 1;
}

static struct ip6t_target ip6t_reject_reg
= { { NULL, NULL }, "REJECT", reject6_target, check, NULL, THIS_MODULE };

static int __init init(void)
{
	if (ip6t_register_target(&ip6t_reject_reg))
		return -EINVAL;
	return 0;
}

static void __exit fini(void)
{
	ip6t_unregister_target(&ip6t_reject_reg);
}

module_init(init);
module_exit(fini);
